/*
 * SIMD functions.  These replace the functions in asm-i386/string.h
 * whenever it makes sense.  These also un-inline those functions.
 *
 * Copyright 1999, Doug Ledford <dledford@redhat.com>
 *
 * These functions are simple and trivial, consider them to be
 * public domain
 */

#include <linux/string.h>
#include <linux/sched.h>
#include <asm/processor.h>
#include <asm/uaccess.h>
#include <asm/i387.h>

extern void * kni_memcpy(void * to, const void * from, size_t n)
{
	unsigned long flags;
	void *ret=to;
	size_t size;
	int recursive = 0;
	char xmm_space[64];

	/*
	 * If the transfer is too small, then use the generic routine.
	 */
	if (n < 128) {
		return(__memcpy(to, from, n));
	}
	kernel_take_fpu_kni(recursive,&xmm_space[0],NULL,flags);

	/*
	 * Align the destination on a 16byte boundary.
	 * The source doesn't have to be aligned.
	 */
	if ( (unsigned long)to & 0xf ) {
		size = 0x10 - ((unsigned long)to & 0xf);
		__asm__ __volatile__("movups (%0),%%xmm0\n\t"
				     "movups %%xmm0,(%1)\n\t"
				     :
				     : "r" (from),
				       "r" (to));
		n -= size;
		from += size;
		to += size;
	}
	/*
	 * If the copy would have tailings, take care of them
	 * now instead of later
	 */
	if(n & 0xf) {
		size = n - 0x10;
		__asm__ __volatile__("movups (%0),%%xmm0\n\t"
				     "movups %%xmm0,(%1)\n\t"
				     :
				     : "r" (from + size),
				       "r" (to + size));
		n &= ~0xf;
	}
	/*
	 * Prefetch the first two cachelines now.
	 */
	__asm__ __volatile__("prefetchnta 0x00(%0)\n\t"
			     "prefetchnta 0x20(%0)\n\t"
			     :
			     : "r" (from));
	/*
	 * Copy 32 bytes at a time.  The single unroll is good
	 * for a 30% performance boost in the copy.  Additional
	 * unrolls are not productive.  We are guaranteed to
	 * have at least 32 bytes of data to copy since the
	 * macro in string.h doesn't call into this function
	 * with less than 64 bytes of copy and we lost < 32
	 * bytes to alignment earlier.
	 */
	while (n >= 0x20) {
		__asm__ __volatile__(
				     "movups 0x00(%0),%%xmm0\n\t"
				     "movups 0x10(%0),%%xmm1\n\t"
				     "movntps %%xmm0,0x00(%1)\n\t"
				     "movntps %%xmm1,0x10(%1)\n\t"
				     : 
				     : "r" (from), "r" (to)
				     : "memory");
		from += 0x20;
		/*
		 * Note: Intermixing the prefetch at *exactly* this point
		 * in time has been shown to be the fastest possible.
		 * Timing these prefetch instructions is a complete black
		 * art with nothing but trial and error showing the way.
		 * To that extent, this optimum version was found by using
		 * a userland version of this routine that we clocked for
		 * lots of runs.  We then fiddled with ordering until we
		 * settled on our highest speen routines.  So, the long
		 * and short of this is, don't mess with instruction ordering
		 * here or suffer permance penalties you will.
		 */
		__asm__ __volatile__(
				     "prefetchnta 0x20(%0)\n\t"
				     : 
				     : "r" (from));
		to += 0x20;
		n -= 0x20;
	}
	if (n) {
		__asm__ __volatile__("movups 0x00(%0),%%xmm0\n\t"
				     "movntps %%xmm0,0x00(%1)\n\t"
				     : 
				     : "r" (from), "r" (to)
				     : "memory");
	}
	SFENCE();
	kernel_release_fpu_kni(recursive,&xmm_space[0],flags);
	return(ret);
}

extern void * kni_memset(void * s, char c, size_t count)
{
	unsigned long flags;
	size_t size;
	void *ret=s;
	int recursive = 0;
	char xmm_space[64];

	/*
	 * If the transfer is too small, then use the generic routine.
	 */
	if (count < 128) {
		return(__memset_generic(s, c, count));
	}
	kernel_take_fpu_kni(recursive,&xmm_space[0],NULL,flags);
	/*
	 * Load up our XMM register with the stuff to set mem with
	 */
	if(c == '\0') {
		__asm__ __volatile__("xorps %%xmm0,%%xmm0\n\t"
				     "movups %%xmm0,(%0)\n\t"
				     :
				     : "r" (s));
	} else {
		__memset_generic(s, c, 0x10);
		__asm__ __volatile__("movups (%0),%%xmm0"
				     :
				     : "r" (s));
	}
	/*
	 * align the destination on a 16 byte boundary, we can simply
	 * do the math to align things since we already populated the
	 * first 16 bytes.
	 */
	size = (0x10 - ((unsigned long)s & 0xf));
	count -= size;
	s += size;
	/*
	 * On the off chance we have tailings due to alignment issues,
	 * do them now to make later more efficient
	 */
	if(count & 0xf) {
		__asm__ __volatile__("movups %%xmm0,(%0)"
				     :
				     : "r" (s + (count - 0x10))
				     : "memory");
		count &= ~0xf;
	}
	/*
	 * Do the copy by plopping out the register to memory.
	 * Note: Unrolling this was *totally* unproductive.  My benchmark
	 * showed that one or two plops per iteration produced the same
	 * speed to within .06 MByte/s of speed.  Considering that the
	 * routine benchmarked at over 3000 MByte/s, .06 is not statistically
	 * significant and only doing one drop per loop simplifies 
	 * overhead of book keeping.
	 */
	while(count) {
		__asm__ __volatile__("movntps %%xmm0,0x00(%0)\n\t"
				     :
				     : "r" (s));
		s += 0x10;
		count -= 0x10;
	}
	SFENCE();
	kernel_release_fpu_kni(recursive,&xmm_space[0],flags);
	return(ret);
}

#define __kni_copy_to_user(to,from,size)				\
do {									\
	int __d0, __d1, tmp, tmp2;					\
	__asm__ __volatile__(						\
		"	movl %1,%4\n"					\
		"	andl $0xf,%4\n"					\
		"	movups (%2),%%xmm0\n"				\
		"1:	movups %%xmm0,(%1)\n"				\
		"	movl $0x10,%3\n"				\
		"	subl %4,%3\n"					\
		"	addl %3,%2\n"					\
		"	addl %3,%1\n"					\
		"	subl %3,%0\n"					\
		"	prefetchnta 0x00(%2)\n"				\
		"	prefetchnta 0x20(%2)\n"				\
		"	jmp 200f\n"					\
		"100:	movups 0x00(%2),%%xmm0\n"			\
		"	movups 0x10(%2),%%xmm1\n"			\
		"2:	movntps %%xmm0,0x00(%1)\n"			\
		"3:	movntps %%xmm1,0x10(%1)\n"			\
		"	addl $0x20,%2\n"				\
		"	prefetchnta 0x20(%2)\n"				\
		"	addl $0x20,%1\n"				\
		"	subl $0x20,%0\n"				\
		"200:	cmpl $0x1f,%0\n"				\
		"	ja 100b\n"					\
		"	cmpl $0xf,%0\n"					\
		"	jbe 300f\n"					\
		"	movups 0x00(%2),%%xmm0\n"			\
		"4:	movntps %%xmm0,0x00(%1)\n"			\
		"	addl $0x10,%2\n"				\
		"	addl $0x10,%1\n"				\
		"	subl $0x10,%0\n"				\
		"300:	testl %0,%0\n"					\
		"	je 400f\n"					\
		"	movl $0x10,%3\n"				\
		"	subl %0,%3\n"					\
		"	subl %3,%1\n"					\
		"	subl %3,%2\n"					\
		"	movups 0x00(%2),%%xmm0\n"			\
		"5:	movups %%xmm0,0x00(%1)\n"			\
		"	addl $0x10,%2\n"				\
		"	addl $0x10,%1\n"				\
		"	xorl %0,%0\n"					\
		"400:\n"						\
		".section .fixup,\"ax\"\n"				\
		"6:	jmp 400b\n"					\
		"7:	addl $0x10,%1\n"				\
		"	addl $0x10,%2\n"				\
		"	subl $0x10,%0\n"				\
		"	jmp 400b\n"					\
		"8:	addl %3,%1\n"					\
		"	addl %3,%2\n"					\
		"	jmp 400b\n"					\
		".previous\n"						\
		".section __ex_table,\"a\"\n"				\
		"	.align 4\n"					\
		"	.long 1b,6b\n"					\
		"	.long 2b,6b\n"					\
		"	.long 3b,7b\n"					\
		"	.long 4b,6b\n"					\
		"	.long 5b,8b\n"					\
		".previous"						\
		: "=&c"(size), "=&D" (__d0), "=&S" (__d1), "=r"(tmp),	\
		  "=r"(tmp2)						\
		: "0"(size), "1"(to), "2"(from)				\
		: "memory");						\
} while (0)

#define __kni_copy_from_user(to,from,size)				\
do {									\
	int __d0, __d1, tmp, tmp2;					\
	__asm__ __volatile__(						\
		"	movl %1,%4\n"					\
		"	andl $0xf,%4\n"					\
		"1:	movups (%2),%%xmm0\n"				\
		"	movups %%xmm0,(%1)\n"				\
		"	movl $0x10,%3\n"				\
		"	subl %4,%3\n"					\
		"	addl %3,%2\n"					\
		"	addl %3,%1\n"					\
		"	subl %3,%0\n"					\
		"	prefetchnta 0x00(%2)\n"				\
		"	prefetchnta 0x20(%2)\n"				\
		"	jmp 100f\n"					\
		"2:	movups 0x00(%2),%%xmm0\n"			\
		"3:	movups 0x10(%2),%%xmm1\n"			\
		"	movntps %%xmm0,0x00(%1)\n"			\
		"	movntps %%xmm1,0x10(%1)\n"			\
		"	addl $0x20,%2\n"				\
		"	prefetchnta 0x20(%2)\n"				\
		"	addl $0x20,%1\n"				\
		"	subl $0x20,%0\n"				\
		"100:	cmpl $0x1f,%0\n"				\
		"	ja 2b\n"					\
		"	cmpl $0xf,%0\n"					\
		"	jbe 200f\n"					\
		"4:	movups 0x00(%2),%%xmm0\n"			\
		"	movntps %%xmm0,0x00(%1)\n"			\
		"	addl $0x10,%2\n"				\
		"	addl $0x10,%1\n"				\
		"	subl $0x10,%0\n"				\
		"200:	testl %0,%0\n"					\
		"	je 300f\n"					\
		"	movl $0x10,%3\n"				\
		"	subl %0,%3\n"					\
		"	subl %3,%1\n"					\
		"	subl %3,%2\n"					\
		"5:	movups 0x00(%2),%%xmm0\n"			\
		"	movups %%xmm0,0x00(%1)\n"			\
		"	addl $0x10,%2\n"				\
		"	addl $0x10,%1\n"				\
		"	xorl %0,%0\n"					\
		"300:\n"						\
		".section .fixup,\"ax\"\n"				\
		"6:	xorps %%xmm0,%%xmm0\n"				\
		"	movups %%xmm0,(%1)\n"				\
		"	movl $0x10,%3\n"				\
		"	subl %4,%3\n"					\
		"	addl %3,%1\n"					\
		"	movl %3,%4\n"					\
		"	movl %0,%3\n"					\
		"	subl %4,%3\n"					\
		"	jmp 600f\n"					\
		"7:	subl $0x10,%0\n"				\
		"	addl $0x10,%1\n"				\
		"400:	movl %0,%3\n"					\
		"	xorps %%xmm0,%%xmm0\n"				\
		"	jmp 600f\n"					\
		"500:	movntps %%xmm0,0x00(%1)\n"			\
		"	movntps %%xmm0,0x10(%1)\n"			\
		"	addl $0x20,%1\n"				\
		"	subl $0x20,%3\n"				\
		"600:	cmpl $0x1f,%3\n"				\
		"	ja 500b\n"					\
		"	cmpl $0xf,%3\n"					\
		"	jbe 700f\n"					\
		"	movntps %%xmm0,0x00(%1)\n"			\
		"	addl $0x10,%1\n"				\
		"	subl $0x10,%3\n"				\
		"700:	testl %3,%3\n"					\
		"	je 300b\n"					\
		"	xorl %4,%4\n"					\
		"	movb %4,(%1)\n"					\
		"	inc %1\n"					\
		"	dec %3\n"					\
		"	jmp 700b\n"					\
		"8:	addl %3,%1\n"					\
		"	movl %0,%3\n"					\
		"	jmp 700b\n"					\
		".previous\n"						\
		".section __ex_table,\"a\"\n"				\
		"	.align 4\n"					\
		"	.long 1b,6b\n"					\
		"	.long 2b,400b\n"				\
		"	.long 3b,7b\n"					\
		"	.long 4b,400b\n"				\
		"	.long 5b,8b\n"					\
		".previous"						\
		: "=&c"(size), "=&D" (__d0), "=&S" (__d1), "=r"(tmp),	\
		  "=q"(tmp2)						\
		: "0"(size), "1"(to), "2"(from)				\
		: "memory");						\
} while (0)


unsigned long
__kni_copy_to_user_nocheck(void *to, const void *from, unsigned long n)
{
	unsigned long flags;
	int recursive = 0;
	char xmm_space[64];
	char xmm_reg_space[64]; /* in case we switch context */

	if (n >= 128) {
		kernel_take_fpu_kni(recursive,&xmm_space[0],&xmm_reg_space[0],flags);
		__kni_copy_to_user(to,from,n);
		SFENCE();
		kernel_release_fpu_kni(recursive,&xmm_space[0],flags);
	} else {
		__copy_user(to,from,n);
	}
	return n;
}

unsigned long
__kni_copy_from_user_nocheck(void *to, const void *from, unsigned long n)
{
	unsigned long flags;
	int recursive = 0;
	char xmm_space[64];
	char xmm_reg_space[64]; /* in case we switch context */

	if (n >= 128) {
		kernel_take_fpu_kni(recursive,&xmm_space[0],&xmm_reg_space[0],flags);
		__kni_copy_from_user(to,from,n);
		SFENCE();
		kernel_release_fpu_kni(recursive,&xmm_space[0],flags);
	} else {
		__copy_user_zeroing(to,from,n);
	}
	return n;
}



unsigned long
kni_copy_to_user(void *to, const void *from, unsigned long n)
{
	unsigned long flags;
	int recursive = 0;
	char xmm_space[64];
	char xmm_reg_space[64]; /* in case we switch context */

	if (access_ok(VERIFY_WRITE, to, n)) {
		if (n >= 128) {
			kernel_take_fpu_kni(recursive,&xmm_space[0],&xmm_reg_space[0],flags);
			__kni_copy_to_user(to,from,n);
			SFENCE();
			kernel_release_fpu_kni(recursive,&xmm_space[0],flags);
		} else {
			__copy_user(to,from,n);
		}
	}
	return n;
}

unsigned long
kni_copy_from_user(void *to, const void *from, unsigned long n)
{
	unsigned long flags;
	int recursive = 0;
	char xmm_space[64];
	char xmm_reg_space[64]; /* in case we switch context */

	if (access_ok(VERIFY_READ, from, n)) {
		if (n >= 128) {
			kernel_take_fpu_kni(recursive,&xmm_space[0],&xmm_reg_space[0],flags);
			__kni_copy_from_user(to,from,n);
			SFENCE();
			kernel_release_fpu_kni(recursive,&xmm_space[0],flags);
		} else {
			__copy_user_zeroing(to,from,n);
		}
	}
	return n;
}


